/******************************************************************************
 *
 *  C++ implementation of a simpified version of fortran routine dgemm of 
 *  Craig C. Douglas, Michael Heroux, Gordon Slishman, and Roger M. Smith by
 *
 *      Enrico Bertolazzi
 *      Dipartimento di Ingegneria Meccanica e Strutturale
 *      Universita` degli Studi di Trento
 *      Via Mesiano 77, I-38050 Trento, Italy
 *      email: enrico.bertolazzi@unitn.it
 *
 * Function:    Strassen-Winograd Matrix-Matrix Multiplication
 *
 * Reference:   Craig C. Douglas, Michael Heroux, Gordon Slishman, and Roger M. Smith,
 *              GEMMW:  A portable Level 3 BLAS Winograd variant of Strassen's  matrix--matrix multiply algorithm,
 *              Yale University Department of Computer Science Report
 *              YALEU/DCS/904, New Haven, CT, 1992 (submitted).
 *
 * Algorithm:   Winograd's variation of Strassen' algorithm, as outlined in
 *              "The Design and Analysis of Computer Algorithms,"
 *              by Aho, Hopcroft and Ullman, 1974, exercise 6.5, page 247.
 *
 * Returns:     C = alpha * op(A) * op(B) + beta * C, where
 *
 *                          { X,              if transX == 'N' or 'n'
 *                  op(X) = { X transpose,    if transX == 'T' or 't'
 *                          { X conj transp,  if transX == 'C' or 'c'
 *                          { X conjugate,    if transX == 'K' or 'k'
 *              and
 *                  alpha and beta are scalars of the same data type as
 *                  the matrices A, B, and C.
 *
 *              The dimensions of the matrices are as follows:
 *
 *                             rows    columns
 *                          +--------+--------+
 *                    op(A) |   M    |    K   |
 *                    op(B) |   K    |    N   |
 *                    C     |   M    |    N   |
 *                          +--------+--------+
 *
 * Naming
 * Convention:  _name, where the underscore is replaced by 
 *                  c   Single precision complex
 *                  d   Double precision real
 *                  s   Single precision real
 *                  z   Double precision complex
 *
 * Modules:     1.  _gemmw is the front end for the computational
 *                  kernel.  _gemmw maps the caller's parameters
 *                  (Fortran perhaps) to the kernel interface, validates
 *                  the parameters, and obtains sufficient auxiliary storage
 *                  (approx. 2/3*N**2 or 5/3*N**2 elements for square
 *                  matrices, as described below), calls the kernel, frees
 *                  auxiliary storage, and returns to the caller.
 *
 *              2.  _gemmb is a Level 3 BLAS header to _gemmw.
 *                  The calling sequence is identical to _gemm and _gemmw will
 *                  automatically allocate memory for its auxiliary space.
 *
 *              3.  _winos is the kernel.  It calls itself and
 *                  subroutines of some linear algebra library, for
 *                  example,
 *                      2A. LAPACK BLAS
 *                      2B. Cray scientific library BLAS
 *                      2C. IBM BLAS for the RISC System 6000
 *                      2D. IBM ESSL BLAS
 *                  The linear algebra subroutines perform ordinary
 *                  matrix and vector operations such as addition,
 *                  subtraction and multiplication of "small" matrices.
 *                  This is an internal routine that is not intended to be
 *                  called directly by the users of _gemmw.
 *
 *              4.  winolap checks to see if the active part of the matrix C
 *                  overlaps with the active part of some other matrix op(X).
 *                  It returns 0 if there is no overlap and 1 if there is.
 *                  This is an internal routine that is not intended to be
 *                  called directly by the users of _gemmw.
 *
 * Storage:     _gemmw obtains auxiliary storage dynamically for the
 *              caller (unless enough is provdied in the aux varaible).
 *              The amount obtained in elements is (rounded up):
 *
 *              (1)    (M*max(K,N) + K*N)/3 + (max(K,N) + M + K + 3N)/2 + 32    
 *
 *              If beta != 0 or either A or B overlaps with C in memory,
 *              then an additional MN amount of storage is needed.
 *
 *              The caller needs to provide adequate virtual storage,
 *              either explicitly (the aux variable) or implicitly
 *              (letting _gemmw do a malloc call itself).  Hence, the
 *              above calculation need never be performed by the user.
 *
 *              Equation (1) is easily derived.  Each level of recursion
 *              needs two major work areas, one of size bounded by
 *
 *                  (M/2+1/2)*(max(K,N)/2+1/2)
 *
 *              and the other bounded by
 *
 *                  (K/2+1/2)*(N/2+1/2).
 *
 *              When M is odd, a vector of length N/2 is required as
 *              well.  The sum is
 *
 *                  (M*max(K,N) + K*N + max(K,N) + M + K + 3N + 2)/4
 *
 *              As the storage requirement decreases by a factor of four
 *              with each level of recursion for the square terms and by
 *              a factor of two for the linear ones and as
 *
 *                  1/4 + 1/16 + 1/256 + ... = 1/3 
 *              and 
 *                  1/2 + 1/4 + 1/8 = 1,
 *
 *              the formula (1) follows trivially.
 *
 * Recursion:   The depth of recursion has no a priori bound applicable
 *              to all problems.  Instead a compilation parameter
 *              "mindim" states the value of min(M,N,K) below which a
 *              conventional matrix-matrix multiplication routine is
 *              invoked.
 *
 * Implementation:  The particular implementation used here follows the
 *                  steps below.  MK and KN are the two large buffers required
 *                  at each level of the recursion.   The quadrants of C
 *                  are used liberally.
 *
 *            Step    MK      C11     C12     C21     C22     KN      Operation
 *            ----    ----    ----    ----    ----    ----    ----    ---------
 *
 *             1.                                             S7      B22-B12
 *                                                            |
 *             2.     S3                                      |       A11-A21
 *                    |                                       |
 *             3.     +---------------------> M4 <------------+       S3*S7
 *                                            |
 *             4.     S1                      |                       A21+A22
 *                    |                       |
 *             5.     |                       |               S5      B12-B11
 *                    |                       |               |
 *             6.     +-----------------------|-----> M5 <----+       S1*S5
 *                    |                       |       |       |
 *             7.     |                       |       |       S6      B22-S5
 *                    |                       |       |       |
 *             8.     S2                      |       |       |       S1-A11
 *                    |                       |       |       |
 *             9.     +-----> M1 <------------|-------|-------+       S2*S6
 *                    |       |               |       |       |
 *            10.     S4      |               |       |       |       A12-S2
 *                    |       |               |       |       |
 *            11.     +-------|-----> M6      |       |       |       S4*B22
 *                            |       |       |       |       |
 *            12.             |       T3 <----|-------+       |       M5+M6
 *                            |       |       |       |       |
 *            13.     M2      |       |       |       |       |       A11*B11
 *                    |       |       |       |       |       |
 *            14.     +-----> T1      |       |       |       |       M1+M2
 *                    |       |       |       |       |       | 
 *            15.     |       +-----> C12     |       |       |       T1+T3
 *                    |       |               |       |       |
 *            16.     |       T2 <------------+       |       |       T1+M4
 *                    |       |                       |       |
 *            17.     |       |                       |       S8      S6-B21
 *                    |       |                       |       |
 *            18.     |       |               M7 <----|-------+       A22*S8
 *                    |       |               |       |
 *            19.     |       +-------------> C21     |               T2-M7
 *                    |       |                       |
 *            20.     |       +---------------------> C22             T2+M5
 *                    |
 *            21.     |       M3                                      A12*B21
 *                    |       |
 *            22.     +-----> C11                                     M2+M3
 *
 *
 * When M, N, or K is odd we get special cases of interest:
 *
 * Odd M/   Conceptually the approach is to imagine matrix A with its
 * odd N    odd middle row duplicated and/or B with its odd middle
 *          column duplicated.  The product would then have a duplicated
 *          middle row or middle column accordingly.  Letting the output
 *          quadrants overlap by one row or column eliminates the
 *          duplicated row and/or column produced from conceptual A and
 *          B.
 *
 * Odd K    Again, imagine B with its odd middle row duplicated, and
 *          imagine A with a column of zeroes inserted after its odd
 *          middle column.  In each operation involving A12 or A22, the
 *          first column is either omitted (because it's zero) or
 *          handled as a virtual column of zeroes.
 *
 * We recurse on conceptual A and B in all cases.
 *
 * Look for "if ( ko )" or "if ( mo )" in two places each in the code to
 * see how the special cases are handled.  There are absolutely no rank
 * one updates in this code.
 *
 *****************************************************************************/

#include "strassen.hh"
#include "alglin.hh"

#include <algorithm>
#include <iostream>

using namespace std ;

namespace strassen {

  /******************************************************************************
   *
   * strassen
   *
   * The recursive routine that does all of the computational work.  The last
   * argument is a pointer into an already allocated chunk of memory.  It is
   * freed by strassen after strasse returns.
   *
   * This is an internal routine that is not intended to be called
   * directly by the users of gemmw.
   *
   *****************************************************************************/
  static
  void
  strassen( indexType const m,          /* Rows of C and A                       */
            indexType const n,          /* Cols of B and C                       */
            indexType const k,          /* Cols of A and Rows of B               */
            valueType const A[],        /* Pointer to coefficients of A          */
            indexType const lda,        /* Leading dimension of A, Fortran style */
            valueType const B[],        /* Pointer to coefficients of B          */
            indexType const ldb,        /* Leading dimension of B, Fortran style */
            valueType       C[],        /* Pointer to coefficients of C          */
            indexType const ldc,        /* Leading dimension of C, Fortran style */
            valueType     * beginValue, /* Pointer to auxillary storage or NULL  */
            valueType     * endValue    /* Number of elements in aux or <= 0     */ ) {

    /******************************************************************\
     * If the matrices are small, then use the classical algorithm.
    \******************************************************************/
    indexType mindim = 800 ;
    if ( m <= mindim || k <= mindim || n <= mindim ) {
      alglin::gemm( "N", "N", m, n, k, 1.0, A, lda, B, ldb, 0.0, C, ldc ) ;
      return ;
    }
 
    /*******************************************************************
     * This is the Winograd variant's of Strassen.
     ******************************************************************/
 
    /*******************************************************************
     * First, determine properties about the 4 submatrices of each of
     *  A, B, and C.
     ******************************************************************/
    indexType mt = m  / 2  ; // truncated m
    indexType mr = m  - mt ; // rounded up m
    indexType mo = mr - mt ; // odd m indicator

    indexType kt = k  / 2  ; // truncated k
    indexType kr = k  - kt ; // rounded up k
    indexType ko = kr - kt ; // odd k indicator

    indexType nt = n  / 2  ; // truncated n
    indexType nr = n  - nt ; // rounded up n
    indexType no = nr - nt ; // odd n indicator

    valueType const * A12 = A   + kt*lda ; // Pointer to upper right submatrix A
    valueType const * A21 = A   + mt     ; // Pointer to lower left  submatrix A
    valueType const * A22 = A12 + mt     ; // Pointer to lower right submatrix A

    valueType const * B12 = B   + nt*ldb ; // Pointer to upper right submatrix B
    valueType const * B21 = B   + kt     ; // Pointer to lower left  submatrix B
    valueType const * B22 = B12 + kt     ; // Pointer to lower right submatrix B
    
    valueType       * C12 = C   + nr*ldc ;
    valueType       * C21 = C   + mr     ;
    valueType       * C22 = C12 + mt     ;

    /*******************************************************************
     * Grab some free storage for the aux matrices and point to the
     * next level's scratch storage.
     ******************************************************************/
    valueType * Wmxk = beginValue ; beginValue += mr * max(kr,nr) ; // Pointer to MK workspace in aux
    valueType * Wkxn = beginValue ; beginValue += kr * nr ;         // Pointer to KN workspace in aux
    
    ASSERT( beginValue+nt < endValue, "bad alloc" ) ;

    /*******************************************************************
     * Finally, do Strassen-Winograd, recursion and all.
     ******************************************************************/
 
    // Wkxn <- S7 = B22 - B12 (kr x nr)
    alglin::gecopy( kr, nr,     B22, ldb,     Wkxn, kr ) ;
    alglin::geadd ( kr, nr, -1, B12, ldb, +1, Wkxn, kr ) ;
 
    // Wmxk <- S3 = A11 - A21 (mt x kr)
    alglin::gecopy( mt, kr,     A+mo,   lda,     Wmxk, mt ) ; // if m odd skip first row 
    alglin::geadd ( mt, kr, -1, A21+mo, lda, +1, Wmxk, mt ) ; // if m odd skip first row 

    // Wmxk = S3 (mt x kr) : Wkxn = S7 (kr x nr)
    // C21 <- M4 = S3 * S7 (mt x nr)
    strassen( mt, nr, kr,
              Wmxk, mt,
              Wkxn, kr,
              C21,  ldc,
              beginValue, endValue ) ;
 
    // Wmxk = S3 (mt x kr) : Wkxn = S7 (kr x nr) : C21 = M4 (mt x nr)
    // Wmxk <- S1 = A21 + A22 (mr x kr)
    if ( ko ) alglin::copy( mr, A21, 1, Wmxk, 1 ) ; // copy A21 column 1 into Wmxk column 1

    alglin::gecopy( mr, kt,     A21+ko*lda, lda,     Wmxk+ko*mr, mr ) ; // if m odd skip first column 
    alglin::geadd ( mr, kt, +1, A22+ko*lda, lda, +1, Wmxk+ko*mr, mr ) ; // if m odd skip first column 
 
    // Wmxk = S1 (mr x kr) : Wkxn = S7 (kr x nr) : C21 = M4 (mt x nr)
    // Wkxn <- S5 = B12 - B11 (kr x nr)
    alglin::gecopy( kr, nr,     B12, ldb,     Wkxn, kr ) ; 
    alglin::geadd ( kr, nr, -1, B,   ldb, +1, Wkxn, kr ) ;

    // Wmxk = S1 (mr x kr) : Wkxn = S5 (kr x nr) : C21 = M4 (mt x nr)
    // C22 <- M5 = S1 * S5 (mr x kr)
    strassen( mr, nt, kr,
              Wmxk,         mr,
              Wkxn + kr*no, kr,
              C22,          ldc,
              beginValue, endValue ) ;

    // Wmxk = S1 (mr x kr) : Wkxn = S5 (kr x nr) : C21 = M4 (mt x nr) : C22 = M5 (mr x kr)
    // Wkxn <- S6 = B22 - S5 (kr x nr)
    alglin::geadd ( kr, nr, +1, B22, ldb, -1, Wkxn, kr ) ;

    // Wmxk = S1 (mr x kr) : Wkxn = S6 (kr x nr) : C21 = M4 (mt x nr) : C22 = M5 (mr x kr)
    // Wmxk <- S2 = S1 - A11 (mr x kr)
    alglin::geadd ( mr, kr, -1, A, lda, +1, Wmxk, mr ) ;

    // Wmxk = S2 (mr x kr) : Wkxn = S6 (kr x nr) : C21 = M4 (mt x nr) : C22 = M5 (mr x kr) 
    // C11 <- M1 = S2 * S6 (mr x nr)
    strassen( mr, nr, kr,
              Wmxk, mr,
              Wkxn, kr,
              C,    ldc,
              beginValue, endValue ) ;

    // Wmxk = S2 (mr x kr) : Wkxn = S6 (kr x nr) : C11 = M1 (mr x nr) : C21 = M4 (mt x nr) : C22 = M5 (mr x kr) 
    // Wmxk <- S4 = A12 - S2 (mr x kt)
    if ( ko ) alglin::scal( mr, -1, Wmxk, 1 ) ; // if common dim is odd, 1st col of A12 = 0 into MK
    alglin::geadd ( mr, kt, +1, A12 + ko*lda, lda, -1, Wmxk + ko*mr, mr ) ;

    // Wmxk = S4 (mr x kt) : Wkxn = S6 (kr x nr) : C11 = M1 (mr x nr) : C21 = M4 (mt x nr) : C22 = M5 (mr x kr) 
    // C12 <- M6 = S4 * B22 (mt x nt)
    valueType * m5 ;
    if ( mo ) {
      m5 = beginValue ; beginValue += nt ;
      alglin::copy( nt, C22, ldc, m5, 1 )  ; // save first row of M5 to aux m5
    }
    strassen( mr, nt, kr,
              Wmxk,       mr, // S4
              B22+no*ldb, ldb,
              C12,        ldc,
              beginValue, endValue ) ;

    // Wmxk = S4 (mr x kt) : Wkxn = S6 (kr x nr) : C11 = M1 (mr x nr) : C12 = M6 (mt x nt) : C21 = M4 (mt x nr) : C22 = M5 (mr x kr) 
    // C12 <- T3 = M5+M6 (mt x nt)
    if ( mo ) {
      alglin::axpy( nt, +1, m5, 1, C12, ldc ) ; // if m odd first row of M6 is 0
      beginValue -= nt ;
    }
    alglin::geadd ( mt, nt, +1, C22+mo, ldc, +1, C12+mo, ldc ) ; // M6 + M5

    // Wmxk = S4 (mr x kt) : Wkxn = S6 (kr x nr) : C11 = M1 (mr x nr) : C12 = T3 (mt x nt) : C21 = M4 (mt x nr) : C22 = M5 (mr x kr) 
    // Wmxk <- M2 = A11 * B11 (mr x nr)
    strassen( mr, nr, kr,
              A,    lda,
              B,    ldb,
              Wmxk, mr,
              beginValue, endValue ) ;

    // Wmxk = M2 (mr x nr) : Wkxn = S6 (kr x nr) : C11 = M1 (mr x nr) : C12 = T3 (mt x nt) : C21 = M4 (mt x nr) : C22 = M5 (mr x kr) 
    // C11 <- T1 = M1 + M2 (mr x nr)
    alglin::geadd ( mr, nr, +1, Wmxk, mr, +1, C, ldc ) ;

    // Wmxk = M2 (mr x nr) : Wkxn = S6 (kr x nr) : C11 = T1 (mr x nr) : C12 = T3 (mt x nt) : C21 = M4 (mt x nr) : C22 = M5 (mr x kr) 
    // C12 = T1 + T3 (done)
    alglin::geadd ( mr, nt, +1, C+no*ldc, ldc, +1, C12, ldc ) ; // if n is odd skip first column of T1

    // Wmxk = M2 (mr x nr) : Wkxn = S6 (kr x nr) : C11 = T1 (mr x nr) : C21 = M4 (mt x nr) : C22 = M5 (mr x kr) 
    // C11 <- T2 = T1 + M4 (mt x nr)
    alglin::geadd ( mt, nr, +1, C21, ldc, +1, C+mo, ldc ) ; // if m is odd skip first row of T1
 
    // Wmxk = M2 (mr x nr) : Wkxn = S6 (kr x nr) : C11 = T2 (mt x nr) : C21 = M4 (mt x nr) : C22 = M5 (mr x kr) 
    // Wkxn <- S8 = S6 - B21 (kr x nr)
    alglin::geadd ( kr, nr, -1, B21, ldb, +1, Wkxn, kr ) ;

    // Wmxk = M2 (mr x nr) : Wkxn = S8 (kr x nr) : C11 = T2 (mt x nr) : C21 = M4 (mt x nr) : C22 = M5 (mr x kr) 
    // C21 <- M7 = A22 * S8 (mt x nr) 
    strassen( mt, nr, kt,
              A22+ko*lda+mo, lda,
              Wkxn+ko,       kr,
              C21,           ldc,
              beginValue, endValue ) ;

    // Wmxk = M2 (mr x nr) : Wkxn = S8 (kr x nr) : C11 = T2 (mt x nr) : C21 = M7 (mt x nr) : C22 = M5 (mr x kr) 
    // C21 = T2 - M7 (done)
    alglin::geadd ( mt, nr, +1, C+mo, ldc, -1, C21, ldc ) ; // if m odd skip first row of T2

    // Wmxk = M2 (mr x nr) : Wkxn = S8 (kr x nr) : C11 = T2 (mt x nr) : C22 = M5 (mr x kr) 
    // C22 = T2 + M5 (done)
    alglin::geadd ( mt, nt, +1, C+no*ldc+mo, ldc, +1, C22+mo, ldc ) ; // if m off skip first row    of T2
                                                                      // if n odd skip first column of T2

    // Wmxk = M2 (mr x nr) : Wkxn = S8 (kr x nr) : C11 = T2 (mt x nr)
    // C11 <- M3 = A12 * B21 (mr x nr)
    strassen( mr, nr, kt,
              A12+ko*lda, lda,
              B21+ko,     ldb,
              C,          ldc,
              beginValue, endValue ) ;

    // Wmxk = M2 (mr x nr) : Wkxn = S8 (kr x nr) : C11 = M3 (mr x nr)
    // C11 = M2 + M3 (done)
    alglin::geadd ( mr, nr, +1, Wmxk, mr, +1, C, ldc ) ;

    /*******************************************************************
     * Return to either _winos or _gemmw.
     ******************************************************************/
    return;
  }

  /******************************************************************************
   *
   * matrixMultiply
   *
   * The front end to the computational kernel _winos.  This checks parameters,
   * allocates memory, does the final alpha*op(A)op(B) + beta*C step, and frees
   * memory.
   *
   *****************************************************************************/

  void
  matrixMultiply( indexType const m,     /* Rows of C and A                       */
                  indexType const n,     /* Cols of B and C                       */
                  indexType const k,     /* Cols of A and Rows of B               */
                  valueType const A[],   /* Pointer to coefficients of A          */
                  indexType const lda,   /* Leading dimension of A, Fortran style */
                  valueType const B[],   /* Pointer to coefficients of B          */
                  indexType const ldb,   /* Leading dimension of B, Fortran style */
                  valueType       C[],   /* Pointer to coefficients of C          */
                  indexType const ldc,   /* Leading dimension of C, Fortran style */
                  valueType       aux[], /* Pointer to auxillary storage or NULL  */
                  indexType       naux   /* Number of elements in aux or <= 0     */ ) {

    /******************************************************************************\
     * Are the matrices really compatible?                                        *
    \******************************************************************************/

    ASSERT ( m > 0 && k > 0 && n > 0,
             "Winograd-Strassen received a bad k=" << k << ", m=" << m << " or n=" << n );

    /******************************************************************************\
     * How much auxillary space is needed and can we use what was passed to us?   *
    \******************************************************************************/

    valueType * beginValue = aux ;
    valueType * endValue   = aux + naux ;

    indexType mx  = max(k,n) ;
    indexType stg = (m * mx + k * n)/3 + (mx + m + k  + 3 * n)/2 + 32 ; // Minimum aux size required for strassen

    if ( naux < stg ) {
      cout << "REALLOCA " << stg << "\n" ;
      beginValue = new valueType[ stg ] ;
      endValue   = beginValue + stg ;
    }
    
    /******************************************************************************\
     * Call strassen to calculate A B                                             *
    \******************************************************************************/
    strassen( m, n, k, A, lda, B, ldb, C, ldc, beginValue, endValue ) ;

    /******************************************************************************\
     * Free any space allocated in matrixMultiply and then return.                *
    \******************************************************************************/
    if ( beginValue != aux ) delete [] beginValue ;

    return;
  }

}
